#! /opt/gnu/bin/perl -s
##

## UsageString is printed by &usage() subroutine
$UsageString= <<'EndUsage';
USAGE: branchlength.pl [-options...] node outfile
    takes an outfile which was generated from a Phylip program
    (unless -usepaup is specified, then outfile from PAUP), and
    makes a table of distances from all taxa to the specified node.
    The default writes the year instead of the name
    node:     name/number of base node
    outfile:  name of file with tree information (output from Phylip)
Options:
    -names:        writes names instead of year 
    -usage:        writes this message and exits
    -verbose:      writes a lot of intermediate info to STDERR
    -usepaup:      indicates that outfile is from PAUP instead of Phylip)
    -noAncestors:  avoids siblings, cousins, and ancestors of the base node
                   that is, only consider base node and its children
    -noZ59:        avoid sequence names Z59
    -xyfile=file.xy:  writes an xyplot file to file.xy
    -blfile=file.bl:  writes list of taxa and distance to base node
    -B=<B>:           computes distance in mutations, assuming B bases
    -Yo=<Yo>:         subtracts offset Yo from year         
** Beware: make sure that no taxa are named with only digits. **
EndUsage

sub usage {
    print $UsageString;
    exit 0;
}

## command line options, and defaults
$usage       ||= 0;
$names       ||= 0;
$verbose     ||= 0;
$usepaup     ||= 0;
$noAncestors ||= 0;
$noZ59       ||= 0;
$B           ||= 0;
$Yo          ||= 0;
$xyfile      ||= "";
$blfile      ||= "";
$norm        ||= 1;   ## normalization factor

&usage if $usage;

# @ARGV are the array of arguements, and is built in 
# "if @ARGV" == "if there are any arguments", shift means grab the first value.
# base will be the node you go back to.

$base = shift @ARGV or &usage;	# get base node number from argument list


## These two hashes specify the phylogenetic tree;
## Each node has a pointer to its parent,
## and has a distance (len) to its parent
%parentOf=();
%lenToParent=();
 
## readPAUP or readPhylip fills the two hashes %parentOf, %lenToParent
## based on the output file from PAUP or Phylip
$usepaup ? &readPAUP : &readPhylip;

sub readPhylip {
    my $prelude=1;
    my $pastbody=0;

    ## read a line from the file, do the test
    while (<>) {
	## skip past the prelude
	## Between\s*And\s*Length is found in Phylip outfile, 
	## just before branching pattern
	$prelude=0 if (/Between\s*And\s*Length/ );
	next if $prelude;
	next if $pastbody;

	## First three words: parent (digital number, left is always a node)
	##                    node   (maybe number, maybe string, all unique)
	##                    length (floating point -- branch length to parent)
	## if the pattern space num space word-or-number space floating point
	## matches then assign the three to the strings $bet, $and, $len
	if (($parent, $node, $len) = /^\s*(\d+)\s+(\S+)\s+([\d.]+)/) {
	    ## The 'node' column provides the unique key 
	    ## (each node has a single parent)
	    $len *= $norm;
	    $parentOf{$node} = $parent;
	    $lenToParent{$node} = $len;
	} else {
	    next if /Between\s*And\s*Length/;
	    next if /^\s*\-+/;
	    next if /^\s*$/;
	    ## if its anything else, then we are done
	    $pastbody=1 if (scalar(%lenToParent) > 0);
	}
    }
}
sub readPAUP {
    my $prelude=1;
    my $pastbody=0;

    ## read a line from the file, do the test
    while (<>) {
	## skip past the prelude
	## "------" is found in PAUP outfile, just before branching pattern
	$prelude=0 if (/^\s*\-+\s*$/);
	next if $prelude;
	next if $pastbody;

	## Two patterns are seen:
	##    LeafName (Number)[*]   NodeNumber  Length
	##    NodeNumber             NodeNumber  Length
        ##    <node>                 <parent>    <len>
	##                    parent (digital number, always a node)
	##                    node   (maybe number, maybe string, all unique)
	##                    len    (floating point -- branch length to parent)
	if ((($node,$parent,$len) = 
	     /^\s*(\S+)\s+\(\d+\)\*?\s+(\d+)\s+([\d.]+(e-?\d+)?)/) ||
	    (($node,$parent,$len) = 
	     /^\s*(\d+)\s+(\d+)\s+([\d.]+(e-?\d+)?)/)) {

	    ## Fill two hashes
	    $len *= $norm;
	    $lenToParent{$node} = $len;
	    $parentOf{$node} = $parent;
	    
	    ## Comment out this next line for debugging
	    printf STDERR "%4d %20s %.5f\n",$parent,$node,$len if $verbose;
	} else {
	    $pastbody=1 if (scalar(%lenToParent) > 0);
	}
    }
}

## At this point we've read in the file, and created two hashes:
## lenToParent{node} is distance between node and parentOf{node}

## Need to get lengths of nodes which are ancestors of $base
## This has has as its keys only the nodes which are direct ancestors
%lenBackToBase = ();
$tmpnode = $base;
$lentotal=0;
print STDERR "Base: $base " if $verbose;
while ($tmpnode) {
    print STDERR "->$tmpnode" if $verbose;
    $lenBackToBase{$tmpnode}=$lentotal;
    $lentotal += $lenToParent{$tmpnode};
    $tmpnode   = $parentOf{$tmpnode}
}
print STDERR "\n";

## Now find lengths to basenode of all nodes

%lenToBase = ();

foreach $node (keys %lenToParent) {
    ## Check to see if the node is in fact a leaf
    ## If it's a plain digital number, then it's not a leaf
    next if $node =~ /^\d+$/;
    next if ($node =~ /^Z59/ && $noZ59);
    print STDERR "$node " if $verbose;
    $tmpnode = $parentOf{$node};
    $lentotal = $lenToParent{$node};
    ## We will follow it back until $tmpnode is either the Base 
    ## Or is a parent of the Base
    while($tmpnode) {
	print STDERR "--> $tmpnode" if $verbose;
	## if $tmpnode is an ancestor of the Base node, then we are done
	## this happens only if lenBackToBase{} has been defined for that node
	if ( defined($tmplen = $lenBackToBase{$tmpnode}) ) {
	    print STDERR "($lentotal+$tmplen)\n" if $verbose;
	    $lentotal += $tmplen;
	    last;
	}
	$lentotal += $lenToParent{$tmpnode};
	print STDERR "($lentotal)" if $verbose;
	$tmpnode = $parentOf{$tmpnode}
    }
    next if ($noAncestors && $tmpnode ne $base);
    print STDERR "OK $node\n" if $verbose;
    $lenToBase{$node} = $lentotal;
}

if ($blfile) {
    open(BL,">$blfile") || die;
    foreach $node (keys %lenToBase) {
	print BL "$node $lenToBase{$node}\n";
    }
    close BL;
}

open(XY,">$xyfile") if $xyfile;
%mval = ( A=>10, B=>20, C=>30, D=>40, J=>50, F=>60, H=>70, E=>80, Z=>90 );
foreach (keys %lenToBase) {
    $dist = $lenToBase{$_};
    $name = $_;
    if (not $names) {
	($c,$yr,$name) = /^(\S)(\d\d)(\S+)$/;
	next unless $c && $yr;

	$dist = int(0.5+$dist*$B) if $B;
	$yr += 1900.5;
	$yr += 100 if $yr < 1950; ## fix y2k bug, leave y2050 bug
	$yr -= $Yo;
	print XY "#e0\n#c $c\n#m $mval{$c}\n","#k $name\n","$yr $dist\n"
	    if $xyfile;
	print "$yr $dist\n";
    } else {
	print "$name $dist\n";
    }
}
    

__END__


    


